package gov.anl.aps.dm.portal.controllers;

import gov.anl.aps.dm.api.ExperimentDsApi;
import gov.anl.aps.dm.common.exceptions.DmException;
import gov.anl.aps.dm.common.exceptions.ObjectAlreadyExists;
import gov.anl.aps.dm.common.exceptions.InvalidRequest;
import gov.anl.aps.dm.portal.model.entities.Experiment;
import gov.anl.aps.dm.portal.model.beans.ExperimentDbFacade;
import gov.anl.aps.dm.portal.model.beans.ExperimentRoleTypeDbFacade;
import gov.anl.aps.dm.portal.model.beans.UserExperimentRoleDbFacade;
import gov.anl.aps.dm.portal.model.entities.ExperimentRoleType;
import gov.anl.aps.dm.portal.model.entities.ExperimentStation;
import gov.anl.aps.dm.portal.model.entities.UserExperimentRole;
import gov.anl.aps.dm.portal.model.entities.UserExperimentRolePK;
import gov.anl.aps.dm.portal.model.entities.UserInfo;
import gov.anl.aps.dm.portal.utilities.DmApiFactory;
import gov.anl.aps.dm.portal.utilities.SessionUtility;
import java.util.List;

import javax.ejb.EJB;
import javax.inject.Named;
import javax.enterprise.context.SessionScoped;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.Converter;
import javax.faces.convert.FacesConverter;
import org.apache.log4j.Logger;

@Named("experimentController")
@SessionScoped
public class ExperimentController extends DmEntityController<Experiment, ExperimentDbFacade> {

    private static final Logger logger = Logger.getLogger(ExperimentController.class.getName());

    @EJB
    private ExperimentDbFacade experimentDbFacade;
    @EJB
    private UserExperimentRoleDbFacade userExperimentRoleDbFacade;
    @EJB
    private ExperimentRoleTypeDbFacade experimentRoleTypeDbFacade;

    private UserExperimentRole currentUserExperimentRole = null;
    private Experiment currentStationExperiment = null;

    private String filterByName = null;
    private String filterByType = null;
    private String filterByStation = null;
    private String filterByDescription = null;
    private String filterByStartDate = null;
    private String filterByEndDate = null;

    private String selectFilterByName = null;
    private String selectFilterByType = null;
    private String selectFilterByStation = null;
    private String selectFilterByDescription = null;
    private String selectFilterByStartDate = null;
    private String selectFilterByEndDate = null;

    public ExperimentController() {
    }

    @Override
    protected ExperimentDbFacade getFacade() {
        return experimentDbFacade;
    }

    @Override
    protected Experiment createEntityInstance() {
        Experiment newExperiment = new Experiment();
        return newExperiment;
    }

    @Override
    public String getEntityTypeName() {
        return "experiment";
    }

    @Override
    public String getCurrentEntityInstanceName() {
        if (getCurrent() != null) {
            return getCurrent().getName();
        }
        return "";
    }

    @Override
    public Experiment findById(Integer id) {
        return experimentDbFacade.findById(id);
    }

    @Override
    public void prepareEntityInsert(Experiment experiment) throws DmException {
        if ((experiment.getName() == null) || (experiment.getName().length() == 0)) {
            throw new InvalidRequest("Experiment name is missing.");
        }
        Experiment existingExperiment = experimentDbFacade.findByName(experiment.getName());
        if (existingExperiment != null) {
            throw new ObjectAlreadyExists("Experiment " + experiment.getName() + " already exists.");
        }
        verifyExperiment(experiment);
        logger.debug("Inserting new experiment " + experiment.getName());
    }

    @Override
    public void prepareEntityUpdate(Experiment experiment) throws DmException {
        if ((experiment.getName() == null) || (experiment.getName().length() == 0)) {
            throw new InvalidRequest("Experiment name is missing.");
        }
        verifyExperiment(experiment);
        logger.debug("Updating experiment " + experiment.getName());
    }

    @Override
    protected String getObjectAlreadyExistMessage(Experiment experiment) {
        if (experiment == null) {
            return null;
        }
        return "Experiment " + experiment.getName() + " already exists.";
    }

    private void verifyExperiment(Experiment experiment) throws DmException {
        if (experiment.getExperimentType() == null) {
            throw new InvalidRequest("Experiment type is missing.");
        }

        if (experiment.getExperimentStation() == null) {
            throw new InvalidRequest("Experiment station is missing.");
        }

        if ((experiment.getStartDate() != null) && (experiment.getEndDate() != null) && (experiment.getEndDate().before(experiment.getStartDate()))) {
            throw new InvalidRequest("Experiment end date is before start date.");
        }
    }

    @Override
    public String update() {
        // Notify DS Web Service
        try {
            ExperimentDsApi api = DmApiFactory.getExperimentDsApi();
            api.updateExperiment(getCurrent().getName());
        } catch (DmException ex) {
            logger.error("Could not notify Data Storage Service: " + ex);
            SessionUtility.addErrorMessage("Error", "Unable to notify Data Storage Service: " + ex.getErrorMessage());
        }
        clear();
        return super.update();
    }

    public UserExperimentRole getCurrentUserExperimentRole() {
        if (currentUserExperimentRole == null) {
            currentUserExperimentRole = new UserExperimentRole();
        }
        return currentUserExperimentRole;
    }

    public void setCurrentUserExperimentRole(UserExperimentRole currentUserExperimentRole) {
        this.currentUserExperimentRole = currentUserExperimentRole;
    }

    public void prepareAddUserExperimentRole(Experiment experiment) {
        UserExperimentRole userExperimentRole = getCurrentUserExperimentRole();
        userExperimentRole.setExperiment(experiment);
    }

    private void addUserExperimentRole() {
        UserInfo user = currentUserExperimentRole.getUserInfo();
        Experiment experiment = currentUserExperimentRole.getExperiment();
        ExperimentRoleType roleType = currentUserExperimentRole.getExperimentRoleType();
        UserExperimentRolePK userExperimentRolePK = new UserExperimentRolePK();
        userExperimentRolePK.setExperimentId(experiment.getId());
        userExperimentRolePK.setUserId(user.getId());
        userExperimentRolePK.setRoleTypeId(roleType.getId());
        currentUserExperimentRole.setUserExperimentRolePK(userExperimentRolePK);

        UserExperimentRole existingRole = userExperimentRoleDbFacade.findByUserAndExperiment(user.getId(),
                experiment.getId());
        if (existingRole != null) {
            SessionUtility.addErrorMessage("Error", "Could not add experiment user: User "
                    + user.getUsername()
                    + " is already associated with experiment "
                    + existingRole.getExperiment().getName() + " as "
                    + existingRole.getExperimentRoleType().getName() + ".");
            return;
        }
        List<UserExperimentRole> userExperimentRoleList = experiment.getUserExperimentRoleList();
        userExperimentRoleList.add(0, currentUserExperimentRole);
        update();
        currentUserExperimentRole = null;
    }

    public void addUserExperimentRolePi() {
        currentUserExperimentRole.setExperimentRoleType(experimentRoleTypeDbFacade.findPiRoleType());
        addUserExperimentRole();
    }

    public void addUserExperimentRoleUser() {
        currentUserExperimentRole.setExperimentRoleType(experimentRoleTypeDbFacade.findUserRoleType());
        addUserExperimentRole();
    }

    public void deleteUserExperimentRole(UserExperimentRole userExperimentRole) {
        Experiment experiment = getCurrent();
        userExperimentRoleDbFacade.remove(userExperimentRole);
        List<UserExperimentRole> userExperimentRoleList = experiment.getUserExperimentRoleList();
        userExperimentRoleList.remove(userExperimentRole);
        updateOnRemoval();
    }

    public Experiment getCurrentStationExperiment() {
        if (currentStationExperiment == null) {
            currentStationExperiment = new Experiment();
        }
        return currentStationExperiment;
    }

    public void setCurrentStationExperiment(Experiment currentStationExperiment) {
        this.currentStationExperiment = currentStationExperiment;
    }

    public void prepareAddStationExperiment(ExperimentStation experimentStation) {
        currentStationExperiment = getCurrentStationExperiment();
        currentStationExperiment.setExperimentStation(experimentStation);
    }

    public void addStationExperiment() {
        try {
            prepareEntityInsert(currentStationExperiment);
            getFacade().create(currentStationExperiment);
            SessionUtility.addInfoMessage("Success", "Created experiment " + currentStationExperiment.getName() + ".");
            resetListDataModel();
            current = currentStationExperiment;
            List<Experiment> stationExperimentList = currentStationExperiment.getExperimentStation().getExperimentList();
            stationExperimentList.add(0, currentStationExperiment);
            currentStationExperiment = null;
        } catch (DmException | RuntimeException ex) {
            SessionUtility.addErrorMessage("Error", "Could not create experiment: " + ex.getMessage());
        }
    }

    public void deleteStationExperiment(Experiment stationExperiment) {
        List<Experiment> stationExperimentList = stationExperiment.getExperimentStation().getExperimentList();
        stationExperimentList.remove(stationExperiment);
        destroy(stationExperiment);
    }

    @Override
    public void clearListFilters() {
        filterByName = null;
        filterByType = null;
        filterByStation = null;
        filterByDescription = null;
        filterByStartDate = null;
        filterByEndDate = null;
    }

    @Override
    public void clearSelectFilters() {
        selectFilterByName = null;
        selectFilterByType = null;
        selectFilterByStation = null;
        selectFilterByDescription = null;
        selectFilterByStartDate = null;
        selectFilterByEndDate = null;
    }

    public String getFilterByName() {
        return filterByName;
    }

    public void setFilterByName(String filterByName) {
        this.filterByName = filterByName;
    }

    public String getFilterByType() {
        return filterByType;
    }

    public void setFilterByType(String filterByType) {
        this.filterByType = filterByType;
    }

    public String getFilterByStation() {
        return filterByStation;
    }

    public void setFilterByStation(String filterByStation) {
        this.filterByStation = filterByStation;
    }

    public String getFilterByDescription() {
        return filterByDescription;
    }

    public void setFilterByDescription(String filterByDescription) {
        this.filterByDescription = filterByDescription;
    }

    public String getFilterByStartDate() {
        return filterByStartDate;
    }

    public void setFilterByStartDate(String filterByStartDate) {
        this.filterByStartDate = filterByStartDate;
    }

    public String getFilterByEndDate() {
        return filterByEndDate;
    }

    public void setFilterByEndDate(String filterByEndDate) {
        this.filterByEndDate = filterByEndDate;
    }

    public String getSelectFilterByName() {
        return selectFilterByName;
    }

    public void setSelectFilterByName(String selectFilterByName) {
        this.selectFilterByName = selectFilterByName;
    }

    public String getSelectFilterByType() {
        return selectFilterByType;
    }

    public void setSelectFilterByType(String selectFilterByType) {
        this.selectFilterByType = selectFilterByType;
    }

    public String getSelectFilterByStation() {
        return selectFilterByStation;
    }

    public void setSelectFilterByStation(String selectFilterByStation) {
        this.selectFilterByStation = selectFilterByStation;
    }

    public String getSelectFilterByDescription() {
        return selectFilterByDescription;
    }

    public void setSelectFilterByDescription(String selectFilterByDescription) {
        this.selectFilterByDescription = selectFilterByDescription;
    }

    public String getSelectFilterByStartDate() {
        return selectFilterByStartDate;
    }

    public void setSelectFilterByStartDate(String selectFilterByStartDate) {
        this.selectFilterByStartDate = selectFilterByStartDate;
    }

    public String getSelectFilterByEndDate() {
        return selectFilterByEndDate;
    }

    public void setSelectFilterByEndDate(String selectFilterByEndDate) {
        this.selectFilterByEndDate = selectFilterByEndDate;
    }

    @FacesConverter(forClass = Experiment.class)
    public static class ExperimentControllerConverter implements Converter {

        @Override
        public Object getAsObject(FacesContext facesContext, UIComponent component, String value) {
            if (value == null || value.length() == 0) {
                return null;
            }
            ExperimentController controller = (ExperimentController) facesContext.getApplication().getELResolver().
                    getValue(facesContext.getELContext(), null, "experimentController");
            return controller.getEntity(getKey(value));
        }

        java.lang.Integer getKey(String value) {
            java.lang.Integer key;
            key = Integer.valueOf(value);
            return key;
        }

        String getStringKey(java.lang.Integer value) {
            StringBuilder sb = new StringBuilder();
            sb.append(value);
            return sb.toString();
        }

        @Override
        public String getAsString(FacesContext facesContext, UIComponent component, Object object) {
            if (object == null) {
                return null;
            }
            if (object instanceof Experiment) {
                Experiment o = (Experiment) object;
                return getStringKey(o.getId());
            } else {
                throw new IllegalArgumentException("Object " + object + " is of type " + object.getClass().getName() + "; expected type: " + Experiment.class.getName());
            }
        }

    }

}
